from functools import partial, wraps
from operator import xor

from .dep import Dep
from .proxy import proxy


class StateModifiedError(Exception):
    """
    Raised when a proxy is modified in a watched (or computed) expression.
    """

    pass


class ReadonlyError(Exception):
    """
    Raised when a readonly proxy is modified.
    """

    pass


def read_trap(method, obj_cls):
    fn = getattr(obj_cls, method)

    @wraps(fn)
    def trap(self, *args, **kwargs):
        if Dep.stack:
            self.proxy_db.attrs(self)["dep"].depend()
        value = fn(self.target, *args, **kwargs)
        if self.shallow:
            return value
        return proxy(value, readonly=self.readonly)

    return trap


def iterate_trap(method, obj_cls):
    fn = getattr(obj_cls, method)

    @wraps(fn)
    def trap(self, *args, **kwargs):
        if Dep.stack:
            self.proxy_db.attrs(self)["dep"].depend()
        iterator = fn(self.target, *args, **kwargs)
        if self.shallow:
            return iterator
        if method == "items":
            return (
                (key, proxy(value, readonly=self.readonly)) for key, value in iterator
            )
        else:
            proxied = partial(proxy, readonly=self.readonly)
            return map(proxied, iterator)

    return trap


def read_key_trap(method, obj_cls):
    fn = getattr(obj_cls, method)

    @wraps(fn)
    def trap(self, *args, **kwargs):
        if Dep.stack:
            key = args[0]
            keydeps = self.proxy_db.attrs(self)["keydep"]
            if key not in keydeps:
                keydeps[key] = Dep()
            keydeps[key].depend()
        value = fn(self.target, *args, **kwargs)
        if self.shallow:
            return value
        return proxy(value, readonly=self.readonly)

    return trap


def write_trap(method, obj_cls):
    fn = getattr(obj_cls, method)

    @wraps(fn)
    def trap(self, *args, **kwargs):
        if Dep.stack:
            raise StateModifiedError()
        old = self.target.copy()
        retval = fn(self.target, *args, **kwargs)
        attrs = self.proxy_db.attrs(self)
        if obj_cls == dict:
            change_detected = False
            keydeps = attrs["keydep"]
            for key, val in self.target.items():
                if old.get(key) is not val:
                    if key in keydeps:
                        keydeps[key].notify()
                    else:
                        keydeps[key] = Dep()
                    change_detected = True
            if change_detected:
                attrs["dep"].notify()
        else:  # list and set
            if self.target != old:
                attrs["dep"].notify()

        return retval

    return trap


def write_key_trap(method, obj_cls):
    fn = getattr(obj_cls, method)
    getitem_fn = getattr(obj_cls, "get")

    @wraps(fn)
    def trap(self, *args, **kwargs):
        if Dep.stack:
            raise StateModifiedError()
        key = args[0]
        attrs = self.proxy_db.attrs(self)
        is_new = key not in attrs["keydep"]
        old_value = getitem_fn(self.target, key) if not is_new else None
        retval = fn(self.target, *args, **kwargs)
        if method == "setdefault" and not self.shallow:
            # This method is only available when readonly is false
            retval = proxy(retval)

        new_value = getitem_fn(self.target, key)
        if is_new:
            attrs["keydep"][key] = Dep()
        if xor(old_value is None, new_value is None) or old_value != new_value:
            attrs["keydep"][key].notify()
            attrs["dep"].notify()
        return retval

    return trap


def delete_trap(method, obj_cls):
    fn = getattr(obj_cls, method)

    @wraps(fn)
    def trap(self, *args, **kwargs):
        if Dep.stack:
            raise StateModifiedError()
        retval = fn(self.target, *args, **kwargs)
        attrs = self.proxy_db.attrs(self)
        attrs["dep"].notify()
        for key in self._orphaned_keydeps():
            attrs["keydep"][key].notify()
            del attrs["keydep"][key]
        return retval

    return trap


def delete_key_trap(method, obj_cls):
    fn = getattr(obj_cls, method)

    @wraps(fn)
    def trap(self, *args, **kwargs):
        if Dep.stack:
            raise StateModifiedError()
        retval = fn(self.target, *args, **kwargs)
        key = args[0]
        attrs = self.proxy_db.attrs(self)
        attrs["dep"].notify()
        attrs["keydep"][key].notify()
        del attrs["keydep"][key]
        return retval

    return trap


trap_map = {
    "READERS": read_trap,
    "KEYREADERS": read_key_trap,
    "ITERATORS": iterate_trap,
    "WRITERS": write_trap,
    "KEYWRITERS": write_key_trap,
    "DELETERS": delete_trap,
    "KEYDELETERS": delete_key_trap,
}


def readonly_trap(method, obj_cls):
    fn = getattr(obj_cls, method)

    @wraps(fn)
    def trap(self, *args, **kwargs):
        raise ReadonlyError()

    return trap


trap_map_readonly = {
    "READERS": read_trap,
    "KEYREADERS": read_key_trap,
    "ITERATORS": iterate_trap,
    "WRITERS": readonly_trap,
    "KEYWRITERS": readonly_trap,
    "DELETERS": readonly_trap,
    "KEYDELETERS": readonly_trap,
}


def construct_methods_traps_dict(obj_cls, traps, trap_map):
    return {
        method: trap_map[trap_type](method, obj_cls)
        for trap_type, methods in traps.items()
        for method in methods
    }



from .proxy import Proxy, TYPE_LOOKUP
from .traps import construct_methods_traps_dict, trap_map, trap_map_readonly


dict_traps = {
    "READERS": {
        "copy",
        "__eq__",
        "__format__",
        "__ge__",
        "__gt__",
        "__le__",
        "__len__",
        "__lt__",
        "__ne__",
        "__repr__",
        "__sizeof__",
        "__str__",
        "keys",
        "__or__",
        "__ror__",
    },
    "KEYREADERS": {
        "get",
        "__contains__",
        "__getitem__",
    },
    "ITERATORS": {
        "items",
        "values",
        "__iter__",
        "__reversed__",
    },
    "WRITERS": {
        "update",
        "__ior__",
    },
    "KEYWRITERS": {
        "setdefault",
        "__setitem__",
    },
    "DELETERS": {
        "clear",
        "popitem",
    },
    "KEYDELETERS": {
        "pop",
        "__delitem__",
    },
}


class DictProxyBase(Proxy):
    def _orphaned_keydeps(self):
        return set(self.proxy_db.attrs(self)["keydep"].keys()) - set(self.target.keys())


def readonly_dict_proxy_init(self, target, shallow=False, **kwargs):
    super(ReadonlyDictProxy, self).__init__(
        target, shallow=shallow, **{**kwargs, "readonly": True}
    )


DictProxy = type(
    "DictProxy",
    (DictProxyBase,),
    construct_methods_traps_dict(dict, dict_traps, trap_map),
)
ReadonlyDictProxy = type(
    "ReadonlyDictProxy",
    (DictProxyBase,),
    {
        "__init__": readonly_dict_proxy_init,
        **construct_methods_traps_dict(dict, dict_traps, trap_map_readonly),
    },
)

TYPE_LOOKUP[dict] = (DictProxy, ReadonlyDictProxy)



import gc
import sys
from weakref import WeakValueDictionary

from .dep import Dep


class ProxyDb:
    """
    Collection of proxies, tracked by the id of the object that they wrap.
    Each time a Proxy is instantiated, it will register itself for the
    wrapped object. And when a Proxy is deleted, then it will unregister.
    When the last proxy that wraps an object is removed, it is uncertain
    what happens to the wrapped object, so in that case the object id is
    removed from the collection.
    """

    def __init__(self):
        self.db = {}
        gc.callbacks.append(self.cleanup)

    def cleanup(self, phase, info):
        """
        Callback for garbage collector to cleanup the db for targets
        that have no other references outside of the db
        """
        # TODO: maybe also run on start? Check performance
        if phase != "stop":
            return

        keys_to_delete = []
        for key, value in self.db.items():
            # Refs:
            # - sys.getrefcount
            # - ref in db item
            if sys.getrefcount(value["target"]) <= 2:
                # We are the last to hold a reference!
                keys_to_delete.append(key)

        for keys in keys_to_delete:
            del self.db[keys]

    def reference(self, proxy):
        """
        Adds a reference to the collection for the wrapped object's id
        """
        obj_id = id(proxy.target)

        if obj_id not in self.db:
            attrs = {
                "dep": Dep(),
            }
            if isinstance(proxy.target, dict):
                attrs["keydep"] = {key: Dep() for key in proxy.target.keys()}
            self.db[obj_id] = {
                "target": proxy.target,
                "attrs": attrs,  # dep, keydep
                # keyed on tuple(readonly, shallow)
                "proxies": WeakValueDictionary(),
            }

        # Use setdefault to put the proxy in the proxies dict. If there
        # was an existing value, it will return that instead. There shouldn't
        # be an existing value, so we can compare the objects to see if we
        # should raise an exception.
        # Seems to be a tiny bit faster than checking beforehand if
        # there is already an existing value in the proxies dict
        result = self.db[obj_id]["proxies"].setdefault(
            (proxy.readonly, proxy.shallow), proxy
        )
        if result is not proxy:
            raise RuntimeError("Proxy with existing configuration already in db")

    def dereference(self, proxy):
        """
        Removes a reference from the database for the given proxy
        """
        obj_id = id(proxy.target)
        if obj_id not in self.db:
            # When there are failing tests, it might happen that proxies
            # are garbage collected at a point where the proxy_db is already
            # cleared. That's why we need this check here.
            # See fixture [clear_proxy_db](/tests/conftest.py:clear_proxy_db)
            # for more info.
            return

        # The given proxy is the last proxy in the WeakValueDictionary,
        # so now is a good moment to see if can remove clean the deps
        # for the target object
        if len(self.db[obj_id]["proxies"]) == 1:
            ref_count = sys.getrefcount(self.db[obj_id]["target"])
            # Ref count is still 3 here because of the reference through proxy.target
            if ref_count <= 3:
                # We are the last to hold a reference!
                del self.db[obj_id]

    def attrs(self, proxy):
        return self.db[id(proxy.target)]["attrs"]

    def get_proxy(self, target, readonly=False, shallow=False):
        """
        Returns a proxy from the collection for the given object and configuration.
        Will return None if there is no proxy for the object's id.
        """
        if id(target) not in self.db:
            return None
        return self.db[id(target)]["proxies"].get((readonly, shallow))


# Create a global proxy collection
proxy_db = ProxyDb()



from .proxy import Proxy, TYPE_LOOKUP
from .traps import construct_methods_traps_dict, trap_map, trap_map_readonly


list_traps = {
    "READERS": {
        "count",
        "index",
        "copy",
        "__add__",
        "__getitem__",
        "__contains__",
        "__eq__",
        "__ge__",
        "__gt__",
        "__le__",
        "__lt__",
        "__mul__",
        "__ne__",
        "__rmul__",
        "__len__",
        "__repr__",
        "__str__",
        "__format__",
        "__sizeof__",
    },
    "ITERATORS": {
        "__iter__",
        "__reversed__",
    },
    "WRITERS": {
        "append",
        "clear",
        "extend",
        "insert",
        "pop",
        "remove",
        "reverse",
        "sort",
        "__setitem__",
        "__delitem__",
        "__iadd__",
        "__imul__",
    },
}


class ListProxyBase(Proxy):
    pass


def readonly_list_proxy_init(self, target, shallow=False, **kwargs):
    super(ReadonlyListProxy, self).__init__(
        target, shallow=shallow, **{**kwargs, "readonly": True}
    )


ListProxy = type(
    "ListProxy",
    (ListProxyBase,),
    construct_methods_traps_dict(list, list_traps, trap_map),
)
ReadonlyListProxy = type(
    "ReadonlyListProxy",
    (ListProxyBase,),
    {
        "__init__": readonly_list_proxy_init,
        **construct_methods_traps_dict(list, list_traps, trap_map_readonly),
    },
)


TYPE_LOOKUP[list] = (ListProxy, ReadonlyListProxy)



__version__ = "0.11.0"


from .proxy import (
    reactive,
    readonly,
    shallow_reactive,
    shallow_readonly,
    to_raw,
)
from .scheduler import scheduler
from .watcher import computed, watch, watch_effect



"""
watchers perform dependency tracking via functions acting on
observable datastructures, and optionally trigger callback when
a change is detected.
"""
from __future__ import annotations

from collections.abc import Container
from functools import partial, wraps
import inspect
from itertools import count
from typing import Any, Callable, Optional, TypeVar
from weakref import ref, WeakSet

from .dep import Dep
from .dict_proxy import DictProxyBase
from .list_proxy import ListProxyBase
from .proxy import Proxy
from .scheduler import scheduler
from .set_proxy import SetProxyBase


T = TypeVar("T", bound=Callable[[], Any])


def watch(
    fn: Callable[[], Any] | Proxy | list[Proxy],
    callback: Optional[Callable] = None,
    sync: bool = False,
    deep: bool | None = None,
    immediate: bool = False,
):
    watcher = Watcher(fn, sync=sync, lazy=False, deep=deep, callback=callback)
    if immediate:
        watcher.dirty = True
        watcher.evaluate()
        if watcher.callback:
            watcher.run_callback(watcher.value, None)
    return watcher


watch_effect = partial(watch, immediate=True, deep=True, callback=None)


def computed(_fn=None, *, deep=True):
    def decorator_computed(fn: T) -> T:
        """
        Create a watcher for an expression.
        Note: make sure fn doesn't need any arguments to run
        and that no reactive state is changed within the expression
        """
        watcher = Watcher(fn, deep=deep)

        @wraps(fn)
        def getter():
            if watcher.dirty:
                watcher.evaluate()
            if Dep.stack:
                watcher.depend()
            return watcher.value

        getter.__watcher__ = watcher
        return getter

    if _fn is None:
        return decorator_computed
    return decorator_computed(_fn)


def traverse(obj, seen=None):
    """
    Recursively traverse the whole tree to make sure
    that all values have been 'get'
    """
    # we are only interested in traversing a fixed set of types
    # otherwise we can just exit
    if isinstance(obj, (dict, DictProxyBase)):
        val_iter = iter(obj.values())
    elif isinstance(obj, (list, ListProxyBase, set, SetProxyBase, tuple)):
        val_iter = iter(obj)
    else:
        return
    # track which objects we have already seen to support(!) full traversal
    # of datastructures with cycles
    # NOTE: a set would provide faster containment checks
    # but these objects are not hashable (except for tuple) because they are mutable
    # converting everything to a hashable thing (because nothing is mutated during
    # traversal) is an option but way more expensive than just using a list
    if seen is None:
        seen = []
    seen.append(obj)
    for v in val_iter:
        if (
            isinstance(
                v, (dict, DictProxyBase, list, ListProxyBase, set, SetProxyBase, tuple)
            )
        ) and v not in seen:
            traverse(v, seen=seen)


# Every Watcher gets a unique ID which is used to
# keep track of the order in which subscribers will
# be notified
_ids = count()


class WrongNumberOfArgumentsError(TypeError):
    """
    Error that is used to signal that the wrong number of arguments is
    used for the callback
    """

    pass


class Watcher:
    def __init__(
        self,
        fn: Callable[[], Any] | Proxy | list[Proxy],
        sync: bool = False,
        lazy: bool = True,
        deep: bool | None = None,
        callback: Callable = None,
    ) -> None:
        """
        sync: Ignore the scheduler
        lazy: Only reevalutate when value is requested
        deep: Deep watch the watched value
        callback: Method to call when value has changed
        """
        self.id = next(_ids)
        if callable(fn):
            if is_bound_method(fn):
                self.fn = weak(fn.__self__, fn.__func__)
            else:
                self.fn = fn
        else:
            self.fn = lambda: fn
            # Default to deep watching when watching a proxy
            # or a list of proxies
            if deep is None:
                deep = True
        self._deps, self._new_deps = WeakSet(), WeakSet()

        self.sync = sync
        if is_bound_method(callback):
            self.callback = weak(callback.__self__, callback.__func__)
        else:
            self.callback = callback
        self.deep = bool(deep)
        self.lazy = lazy
        self.dirty = self.lazy
        self.value = None if self.lazy else self.get()
        self._number_of_callback_args = None

    def update(self) -> None:
        if self.lazy:
            self.dirty = True
        elif self.sync:
            self.run()
        else:
            scheduler.queue(self)

    def evaluate(self) -> None:
        self.value = self.get()
        self.dirty = False

    def run(self) -> None:
        """Called by scheduler"""
        value = self.get()
        if self.deep or isinstance(value, Container) or value != self.value:
            old_value = self.value
            self.value = value
            if self.callback:
                self.run_callback(self.value, old_value)

    def run_callback(self, new, old) -> None:
        """
        Runs the callback. When the number of arguments is still unknown
        for the callback, it will fall into the try/except contstruct
        to figure out the right number of arguments.
        After running the callback one time, the number of arguments
        is known and the callback can be called with the correct
        amount of arguments.
        """
        if self._number_of_callback_args == 1:
            self.callback(new)
            return
        if self._number_of_callback_args == 2:
            self.callback(new, old)
            return
        if self._number_of_callback_args == 0:
            self.callback()
            return

        try:
            self._run_callback(new)
            self._number_of_callback_args = 1
            return
        except WrongNumberOfArgumentsError:
            pass

        try:
            self._run_callback()
            self._number_of_callback_args = 0
            return
        except WrongNumberOfArgumentsError:
            pass

        self._run_callback(new, old)
        self._number_of_callback_args = 2

    def _run_callback(self, *args) -> None:
        """
        Run the callback with the given arguments. When the callback
        raises a TypeError, check to see if the error results from
        within the callback or from calling the callback with the
        wrong number of arguments.
        Raises WrongNumberOfArgumentsError if callback was called
        with the wrong number of arguments.
        """
        try:
            self.callback(*args)
        except TypeError as e:
            frames = inspect.trace()
            try:
                if len(frames) != 1:
                    raise
                raise WrongNumberOfArgumentsError(str(e)) from e
            finally:
                del frames

    def get(self) -> Any:
        Dep.stack.append(self)
        try:
            value = self.fn()
            if self.deep:
                traverse(value)
        finally:
            Dep.stack.pop()
            self.cleanup_deps()
        return value

    def add_dep(self, dep: Dep) -> None:
        if dep not in self._new_deps:
            self._new_deps.add(dep)
            if dep not in self._deps:
                dep.add_sub(self)

    def cleanup_deps(self) -> None:
        for dep in self._deps:
            if dep not in self._new_deps:
                dep.remove_sub(self)
        self._deps, self._new_deps = self._new_deps, self._deps
        self._new_deps.clear()

    def depend(self) -> None:
        """This function is used by other watchers to depend on everything
        this watcher depends on."""
        if Dep.stack:
            for dep in self._deps:
                dep.depend()

    @property
    def fn_fqn(self) -> str:
        return f"{self.fn.__module__}.{self.fn.__qualname__}"


def weak(obj: Any, method: Callable):
    """
    Returns a wrapper for the given method that will only call the method if the
    given object is not garbage collected yet. It does so by using a weakref.ref
    and checking its value before calling the actual method when the wrapper is
    called.
    """
    weak_obj = ref(obj)

    sig = inspect.signature(method)
    nr_arguments = len(sig.parameters)

    if nr_arguments == 1:

        @wraps(method)
        def wrapped():
            if this := weak_obj():
                return method(this)

        return wrapped
    elif nr_arguments == 2:

        @wraps(method)
        def wrapped(new):
            if this := weak_obj():
                return method(this, new)

        return wrapped
    elif nr_arguments == 3:

        @wraps(method)
        def wrapped(new, old):
            if this := weak_obj():
                return method(this, new, old)

        return wrapped
    else:
        raise WrongNumberOfArgumentsError(
            "Please use 1, 2 or 3 arguments for callbacks"
        )


def is_bound_method(fn: Callable):
    """
    Returns whether the given function is a bound method.
    """
    return hasattr(fn, "__self__") and hasattr(fn, "__func__")



from functools import partial

from .proxy_db import proxy_db


class Proxy:
    """
    Proxy for an object/target.

    Instantiating a Proxy will add a reference to the global proxy_db and
    destroying a Proxy will remove that reference.

    Please use the `proxy` method to get a proxy for a certain object instead
    of directly creating one yourself. The `proxy` method will either create
    or return an existing proxy and makes sure that the db stays consistent.
    """

    __hash__ = None
    __slots__ = ["target", "readonly", "shallow", "proxy_db", "__weakref__"]

    def __init__(self, target, readonly=False, shallow=False):
        self.target = target
        self.readonly = readonly
        self.shallow = shallow
        self.proxy_db = proxy_db
        self.proxy_db.reference(self)

    def __del__(self):
        self.proxy_db.dereference(self)


# Lookup dict for mapping a type (dict, list, set) to a method
# that will convert an object of that type to a proxied version
TYPE_LOOKUP = {}


def proxy(target, readonly=False, shallow=False):
    """
    Returns a Proxy for the given object. If a proxy for the given
    configuration already exists, it will return that instead of
    creating a new one.

    Please be aware: this only works on plain data types: dict, list,
    set and tuple!
    """
    # The object may be a proxy already, so check if it matches the
    # given configuration (readonly and shallow)
    if isinstance(target, Proxy):
        if readonly == target.readonly and shallow == target.shallow:
            return target
        else:
            # If the configuration does not match,
            # unwrap the target from the proxy so that the right
            # kind of proxy can be returned in the next part of
            # this function
            target = target.target

    # Note that at this point, target is always a non-proxy object
    # Check the proxy_db to see if there's already a proxy for the target object
    existing_proxy = proxy_db.get_proxy(target, readonly=readonly, shallow=shallow)
    if existing_proxy is not None:
        return existing_proxy

    # We can only wrap the following datatypes
    if not isinstance(target, (dict, list, tuple, set)):
        return target

    # Otherwise, create a new proxy
    proxy_type = None

    for target_type, (writable_proxy_type, readonly_proxy_type) in TYPE_LOOKUP.items():
        if isinstance(target, target_type):
            proxy_type = readonly_proxy_type if readonly else writable_proxy_type
            break
    else:
        if isinstance(target, tuple):
            return tuple(proxy(x, readonly=readonly, shallow=shallow) for x in target)

    return proxy_type(target, readonly=readonly, shallow=shallow)


reactive = proxy
readonly = partial(proxy, readonly=True)
shallow_reactive = partial(proxy, shallow=True)
shallow_readonly = partial(proxy, shallow=True, readonly=True)


def to_raw(target):
    """
    Returns a raw object from which any trace of proxy has been replaced
    with its wrapped target value.
    """
    if isinstance(target, Proxy):
        return to_raw(target.target)

    if isinstance(target, list):
        return [to_raw(t) for t in target]

    if isinstance(target, dict):
        return {key: to_raw(value) for key, value in target.items()}

    if isinstance(target, tuple):
        return tuple(to_raw(t) for t in target)

    if isinstance(target, set):
        return {to_raw(t) for t in target}

    return target



from .proxy import Proxy, TYPE_LOOKUP
from .traps import construct_methods_traps_dict, trap_map, trap_map_readonly


set_traps = {
    "READERS": {
        "copy",
        "difference",
        "intersection",
        "isdisjoint",
        "issubset",
        "issuperset",
        "symmetric_difference",
        "union",
        "__and__",
        "__contains__",
        "__eq__",
        "__format__",
        "__ge__",
        "__gt__",
        "__iand__",
        "__ior__",
        "__isub__",
        "__ixor__",
        "__le__",
        "__len__",
        "__lt__",
        "__ne__",
        "__or__",
        "__rand__",
        "__repr__",
        "__ror__",
        "__rsub__",
        "__rxor__",
        "__sizeof__",
        "__str__",
        "__sub__",
        "__xor__",
    },
    "ITERATORS": {
        "__iter__",
    },
    "WRITERS": {
        "add",
        "clear",
        "difference_update",
        "intersection_update",
        "discard",
        "pop",
        "remove",
        "symmetric_difference_update",
        "update",
    },
}


class SetProxyBase(Proxy):
    pass


def readonly_set_proxy_init(self, target, shallow=False, **kwargs):
    super(ReadonlySetProxy, self).__init__(
        target, shallow=shallow, **{**kwargs, "readonly": True}
    )


SetProxy = type(
    "SetProxy", (SetProxyBase,), construct_methods_traps_dict(set, set_traps, trap_map)
)
ReadonlySetProxy = type(
    "ReadonlysetProxy",
    (SetProxyBase,),
    {
        "__init__": readonly_set_proxy_init,
        **construct_methods_traps_dict(set, set_traps, trap_map_readonly),
    },
)

TYPE_LOOKUP[set] = (SetProxy, ReadonlySetProxy)
